/*
** ARGOBOT
** Jess Cea Avin - jcea@argo.es
**
** 05/May/97 Versin 1.00  Primera versin del parser
** 09/May/97 Corregidos diversos problemas con la segmentacin
**           de las mscaras NICK.
**
** 12/May/97 Versin 1.10
**           Aadidos los canales protegidos y los comandos con clave.
**
** 16/May/97 Versin 1.11
**           No imprime informacin de debug al parsear el fichero de config.
**
** 19/May/97 Versin 1.30
**           Aadidos temporizadores para acciones peridicas
**           Aadidas comprobaciones de rebosamiento al leer la configuracin
**
** 30/May/97 Versin 1.31
**           Cuando se inicia un nuevo parsing del fichero de configuracin,
**           pone todas variables globales a cero, por lo que pudiera pasar.
**
** 13/Jun/97 Versin 1.32
**           Soporte para canales locales.
**
** 07/Ago/97 Versin 1.4
**           Aadidos notices cuando un usuario entra en el canal.
**
** 06/Ene/98 Versin 1.5
**           allow_any_ban.
**
** 07/Jul/98 Version 1.6
**           passwd, case insensitive, "argobot.conf", canales sin modos.
**
** 08/Jul/98 Version 1.7
**           Se usan las nuevas funciones "strcasecmp2()" y "strncasecmp2()" porque
**           NT no tiene "strcasecmp()" ni "strncasecmp()" y asi, ademas, se realiza
**           la equivalencia de nicks.
**           Tambien se usan esas funciones para leer el fichero de configuracion,
**           aunque aqui las unicas equivalencias seran las mayusculas/minusculas.
*/

#define MAX_CONFIG	(64ul*BUF_LEN)

char config[MAX_CONFIG];
unsigned long config_pos;

int conf;	/* Handle del fichero de configuracin */
int linea;	/* Linea del fichero de configuracin que se est procesando */

typedef struct {
	enum {T_EOF,T_EOL,T_LITERAL,
		T_IRCNICK,T_IRCPASSWD,
		T_NICK,T_SERVER,
		T_PORT,T_AWAY,
		T_NICK_COLLIDE,
		T_TIMEOUT_KLINES,
		T_UMBRAL_KLINE,
		T_JOIN,T_LOG,
		T_TIMEOUT_WHOWAS,
		T_ON,T_IF,T_IN,T_THEN,
		T_SET,T_GROUP,T_MODE,
		T_EXPN_NICK,
		T_AUTOOP,T_PRIVIL,
		T_MODE_DEFAULT,
		T_MODE_DISALLOW,
		T_TIMER,T_NOTICE,
		T_ALLOW_ANY_BAN,
		T_PASSWD
	} tipo;
	char literal[BUF_LEN];
} token_t;

void get_token(int conf,token_t *token)
{
char buf[BUF_LEN];
char *p=buf;
ssize_t i;
static int eol=0;

	if(eol) {
		linea++;
		eol=0;
		token->tipo=T_EOL;
		return;
	}
	while(!0) {
		i=read(conf,p,1);
		if(i==1) {
			if((*p=='\t')||(*p==' ')||(*p=='\n')) {
				if(*p=='\n') eol=1;	/* Lo recuerda */
				*p='\0';
				if(p!=buf) break;	/* Tenemos un Token */
				if(eol) {
					linea++;
					eol=0;
					token->tipo=T_EOL;
					return;
				}
			} else {	/* El else por si principio linea */
				if(*p=='#') break;
				p++;
			}
		} else if(!i) {	/* Final de fichero */
			*p='\0';
			if(p!=buf) break;
			token->tipo=T_EOF;
			return;
		} else {
			perror("Error leyendo el fichero de configuracin"),
			exit(errno);
		}		
	}
	if(buf[0]=='#') {	/* Lee hasta el final de la linea */
		linea++;
		p=buf;
		while(read(conf,p,1)>0) if(*p=='\n') break;
		token->tipo=T_EOL;
		return;
	}
	token->tipo=T_EOF;	/* Comodn para abreviar lo siguiente */
	if(!strcasecmp2(buf,"IRCnick")) token->tipo=T_IRCNICK;
	else if(!strcasecmp2(buf,"IRCpasswd")) token->tipo=T_IRCPASSWD;
	else if(!strcasecmp2(buf,"nick")) token->tipo=T_NICK;
	else if(!strcasecmp2(buf,"server")) token->tipo=T_SERVER;
	else if(!strcasecmp2(buf,"port")) token->tipo=T_PORT;
	else if(!strcasecmp2(buf,"away")) token->tipo=T_AWAY;
	else if(!strcasecmp2(buf,"nick_collide")) token->tipo=T_NICK_COLLIDE;
	else if(!strcasecmp2(buf,"timeout_klines")) token->tipo=T_TIMEOUT_KLINES;
	else if(!strcasecmp2(buf,"umbral_kline")) token->tipo=T_UMBRAL_KLINE;
	else if(!strcasecmp2(buf,"timeout_whowas")) token->tipo=T_TIMEOUT_WHOWAS;
	else if(!strcasecmp2(buf,"join")) token->tipo=T_JOIN;
	else if(!strcasecmp2(buf,"log")) token->tipo=T_LOG;
	else if(!strcasecmp2(buf,"on")) token->tipo=T_ON;
	else if(!strcasecmp2(buf,"if")) token->tipo=T_IF;
	else if(!strcasecmp2(buf,"in")) token->tipo=T_IN;
	else if(!strcasecmp2(buf,"then")) token->tipo=T_THEN;
	else if(!strcasecmp2(buf,"set")) token->tipo=T_SET;
	else if(!strcasecmp2(buf,"group")) token->tipo=T_GROUP;
	else if(!strcasecmp2(buf,"mode")) token->tipo=T_MODE;
	else if(!strcasecmp2(buf,"$nick")) token->tipo=T_EXPN_NICK;
	else if(!strcasecmp2(buf,"autoop")) token->tipo=T_AUTOOP;
	else if(!strcasecmp2(buf,"privil")) token->tipo=T_PRIVIL;
	else if(!strcasecmp2(buf,"mode_default")) token->tipo=T_MODE_DEFAULT;
	else if(!strcasecmp2(buf,"mode_disallow")) token->tipo=T_MODE_DISALLOW;
	else if(!strcasecmp2(buf,"timer")) token->tipo=T_TIMER;
	else if(!strcasecmp2(buf,"notice")) token->tipo=T_NOTICE;
	else if(!strcasecmp2(buf,"allow_any_ban")) token->tipo=T_ALLOW_ANY_BAN;
	else if(!strcasecmp2(buf,"passwd")) token->tipo=T_PASSWD;
	else {
		token->tipo=T_LITERAL;
		strcpy(token->literal,buf);
	}
	return;
}

void p_error(char *msg)
{
	fprintf(stderr,"%s. Lnea %d\n",msg,linea);
	exit(-1);
}

void sintaxis(void)
{
	p_error("Error de sintaxis");
}

void config_overflow(void)
{
	fprintf(stderr,"Fichero de configuracin demasiado largo."
		" Lnea %d\n",linea);
	exit(-1);
}

void get_opt_param(char **pos)
{
token_t token;

	*pos=NULL;
	get_token(conf,&token);
	if((token.tipo==T_EOF)||(token.tipo==T_EOL)) return;
	if(token.tipo!=T_LITERAL) sintaxis();
	*pos=config+config_pos;
	strcpy(*pos,token.literal);
	config_pos+=1+strlen(*pos);
	if(config_pos>=MAX_CONFIG) config_overflow();
}

void get_param(char **pos)
{
token_t token;

	get_token(conf,&token);
	if(token.tipo!=T_LITERAL) sintaxis();
	*pos=config+config_pos;
	strcpy(*pos,token.literal);
	config_pos+=1+strlen(*pos);
	if(config_pos>=MAX_CONFIG) config_overflow();
}

void get_final(char **pos)
{
token_t token;

	get_param(pos);
	get_token(conf,&token);
	if((token.tipo!=T_EOL)&&(token.tipo!=T_EOF)) sintaxis();	
}

void fin_linea()
{
token_t token;

	get_token(conf,&token);
	if((token.tipo!=T_EOL)&&(token.tipo!=T_EOF)) sintaxis();	
}

void get_numero(unsigned long *numero)
{
char	*p;

	get_param(&p);
	*numero=atol(p);
	while((*p>='0')&&(*p<='9')) p++;
	if(*p!='\0') sintaxis();
}

void get_numero_final(unsigned long *numero)
{
char *p;

	get_final(&p);
	*numero=atol(p);
	while((*p>='0')&&(*p<='9')) p++;
	if(*p!='\0') sintaxis();
}

void get_texto_final(char **texto)
{
char *p;

	*texto=p=config+config_pos;
	while(read(conf,p,1)>0)
		if((*p!=' ')&&(*p!='\t')) break;
	
	do {	/* Miramos la barra de escape */
		if(*p=='\\') {
			read(conf,p,1);	/* Nos saltamos el fin de Lnea */
			linea++;
			while(read(conf,p,1)>0) {	/* Nos saltamos cosas */
				if((*p!=' ')&&(*p!='\t')) break;
			}
		}
		if(*p=='\n') break;
		p++;
	} while(read(conf,p,1)>0);
	linea++;
	*p='\0';
	config_pos+=1+strlen(*texto);
	if(config_pos>=MAX_CONFIG) config_overflow();
}

void segmentacion(nicks_t *nick,char *mascara)
{
char *p,*p2;

	nick->nick=nick->usuario=nick->dominio=NULL;
	p=mascara;
/* Segmentacin en nick!usuario@dominio */
	p2=strchr(p,'!');
	if(p2==NULL) sintaxis();
	*p2='\0';
	if(*p!='*') nick->nick=p;
	p=++p2;
	p2=strchr(p2,'@');
	if(p2==NULL) sintaxis();
	*p2='\0';
	if(*p!='*') nick->usuario=p;
	if(*(++p2)!='*') nick->dominio=p2;	
}

void lee_config(void)
{
token_t token;

	inicializa_global();
	config_pos=0;
	linea=1;

	conf=open(fichero_config,O_RDONLY);
	if(conf==-1) {
		perror("No se puede abrir el fichero de configuracin");
		exit(errno);
	}
	do {
		get_token(conf,&token);
		
		if(token.tipo==T_EOF) break;	/* Hemos terminado */
		if(token.tipo==T_ALLOW_ANY_BAN) {
			canales[num_canales-1].allow_any_ban=!0;
			fin_linea();
		}
		if(token.tipo==T_LITERAL) sintaxis();
		if(token.tipo==T_EOL) continue;
		if(token.tipo==T_IRCNICK) get_final(&IRCop_nick);
		else if(token.tipo==T_IRCPASSWD) get_final(&IRCop_clave);
		else if(token.tipo==T_NICK) get_final(&Nick);
		else if(token.tipo==T_SERVER) get_final(&IRCop_servidor);
		else if(token.tipo==T_PORT) get_numero_final(&IRCop_puerto);
		else if(token.tipo==T_AWAY) get_texto_final(&IRCop_away);
		else if(token.tipo==T_NICK_COLLIDE)
			get_texto_final(&IRCop_collide);
		else if(token.tipo==T_NOTICE)
			get_texto_final(&canales[num_canales-1].notice);
		else if(token.tipo==T_TIMEOUT_KLINES)
			get_numero_final(&TIMEOUT_KLINES);
		else if(token.tipo==T_UMBRAL_KLINE)
			get_numero_final(&UMBRAL_KLINE);
		else if(token.tipo==T_TIMEOUT_WHOWAS)
			get_numero_final(&TIMEOUT_WHOWAS);
		else if(token.tipo==T_MODE_DEFAULT)
			get_texto_final(&canales[num_canales-1].mode_default);
		else if(token.tipo==T_MODE_DISALLOW)
			get_texto_final(&canales[num_canales-1].mode_disallow);			
		else if(token.tipo==T_PASSWD)
			get_texto_final(&Passwd);	
		else if(token.tipo==T_AUTOOP) {
		int i;
		char *p;
		
			get_final(&p);

			for(i=0;i<num_grupos;i++) {
				if(!strcmp(p,grupos[i].nombre)) {
					canales[num_canales-1].autoop=i;
					break;
				}
			}
			if(i==num_grupos) p_error("Grupo Desconocido");
		} else if(token.tipo==T_TIMER) {
		unsigned long tiempo;
		
			get_numero(&tiempo);
			timers[num_timers].siguiente=time(NULL)+(time_t)tiempo;
			get_numero(&tiempo);
			timers[num_timers].intervalo=(time_t)tiempo;
			get_texto_final(&(timers[num_timers++].accion));
			if(num_timers>=MAX_NUM_TIMERS)
				p_error("Demasiados timers");
		} else if(token.tipo==T_PRIVIL) {
		int i;
		char *p;
		
			get_final(&p);

			for(i=0;i<num_grupos;i++) {
				if(!strcmp(p,grupos[i].nombre)) {
					canales[num_canales-1].privil=i;
					break;
				}
			}
			if(i==num_grupos) p_error("Grupo Desconocido");		
		} else if(token.tipo==T_SET) {
		char *p;
		nicks_t *m=&nicks[num_nicks++];

			if(num_nicks>=MAX_NUM_NICKS)
				p_error("Demasiados Alias");		
			get_param(&(m->nombre));	/* Nombre del alias */
			get_param(&p);	/* Mscara */
			segmentacion(m,p);
			get_opt_param(&(m->clave));
		} else if(token.tipo==T_GROUP) {
		char *p;
		int i,j;
		grupo_t *g=&grupos[num_grupos];
		
			if(num_grupos>=MAX_NUM_GRUPOS)
				p_error("Demasiados grupos");		
			get_param(&(g->nombre));
			g->num_nicks=0;
			get_param(&p);
			do {
				for(i=0;i<num_grupos-1;i++) {
					if(!strcmp(p,grupos[i].nombre)) {
						for(j=0;i<grupos[i].num_nicks;j++) {
							g->nicks[g->num_nicks++]=
								grupos[i].nicks[j];
						}
					}
				}
				for(i=0;i<num_nicks;i++) {
					if(!strcmp(p,nicks[i].nombre)) {
						if(g->num_nicks>=MAX_NUM_NICKS_GRUPO)
							p_error("Grupo demasiado grande");
						g->nicks[g->num_nicks++]=i;
						break;
					}
				}
				if(!g->num_nicks) p_error("Alias desconocido");
				get_opt_param(&p);
			} while(p!=NULL);
			num_grupos++;		
		} else if(token.tipo==T_JOIN) {
		char *p;
		
			canales[num_canales].nombre=config+config_pos;
			config[config_pos++]='#';
			get_param(&p);	/* Lee el canal */
			if((*p=='&')||(*p=='+'))	/* Canal local  o sin modos */
				canales[num_canales].nombre++; /* Salta el # */
			get_opt_param(&canales[num_canales].clave);
			canales[num_canales].log=NULL;
			canales[num_canales].mode_default=NULL;
			canales[num_canales].mode_disallow=NULL;
			canales[num_canales].notice=NULL;
			canales[num_canales].allow_any_ban=0;	/* NO */
			canales[num_canales].autoop=-1;
			canales[num_canales++].privil=-1;
			if(num_canales>=MAX_NUM_CANALES)
				p_error("Demasiados canales");
		} else if(token.tipo==T_LOG)
			get_final(&canales[num_canales-1].log);
						
	} while(!0);
}
